#include <sys/param.h>
#include <sys/sysctl.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "netbsd_emu.h"

/* Both sh and make seems to use getopt after the invocation of
 * execve et al. in case it return.  This will not return the expected
 * result if we have messed with getopt too, so use our own version. */
int my_getopt(int nargc, char * const nargv[], const char *ostr);
#include "getopt.c"

extern char **environ;
static FILE* tfile = stderr;

/* The maximum number or elements we allocate in our argv arrays. */
#define ARGMAX 6000
#define ENVMAX 6000

const char* emul_hw_machine = EMUL_HW_MACHINE;

static void
xfatal(const char* msg)
{
  fprintf(tfile, "%s", msg);
  exit(1);
}


static int
xasprintf(char **ret, const char *format, ...) 
  __attribute__((__format__(__printf__, 2, 3)));

static int
xasprintf(char **ret, const char *format, ...)
{
  int status;
  va_list args;

  va_start(args, format);
  status = vasprintf(ret, format, args);
  va_end(args);

  if (*ret == NULL)
    xfatal("No memory!\n");

  return status;
}


static char*
xstrdup(const char* s)
{
  void *p = strdup(s);
  if (p == NULL)
    xfatal("No memory!\n");

  return p;
}


static void*
xmalloc(size_t size)
{
  void *p = malloc(size);
  if (p == NULL)
    xfatal("No memory!\n");

  return p;
}


char*
get_emul_machine_arch(const char *format)
{
  static char *s = NULL;

  free(s);
  xasprintf(&s, format, emul_hw_machine);

  return s;
}


/* Convert a path in the host's file system, to the path the target will
 * see. */
const char*
host_to_target_path(const char* path)
{
  size_t pathlen;
  const char *p = path;

  pathlen = strlen(FILESYSTEM_PATH);
  if (!strncmp(path, FILESYSTEM_PATH, pathlen))
  {
    if (path[pathlen] == '\0')
      p = "/";
    else if (path[pathlen] == '/')
      p += pathlen;
  }

  return p;
}


/* Returns true if the path would not be changed by calling
 * target_to_host_path() */
int
is_common_path(const char* path)
{
  /* The path does not need to be changed if it is a relative path, or
   * if it is in one of the directories that has the same path in the
   * host and target file systems. */
  int is_correct_path =
    path[0] != '/'
    || (!strncmp("/dev", path, 4) && (path[4] == '\0' || path[4] == '/'))
    || (!strncmp("/tmp", path, 4) && (path[4] == '\0' || path[4] == '/'))
    || (!strncmp(COMMON_PATH, path, strlen(COMMON_PATH))
        && (path[strlen(COMMON_PATH)] == '\0'
            || path[strlen(COMMON_PATH)] == '/'))
    /* The FILESYSTEM_PATH shoul never be seen by the target, but it can
     * happen, unfortunately.  For example, the perl package grep in
     * output from "gcc -E" to find dependencies (and we do not want
     * to run gcc emulated in order to prevent this...) */
    || (!strncmp(FILESYSTEM_PATH, path, strlen(FILESYSTEM_PATH))
        && (path[strlen(FILESYSTEM_PATH)] == '\0'
            || path[strlen(FILESYSTEM_PATH)] == '/'));

  return is_correct_path;
}


/* Convert a path, as seen by the target, to the path to the actual file
 * in the host's file system. */
char*
target_to_host_path(const char* path)
{
  static char *new_path = NULL;

  free(new_path);
  xasprintf(&new_path,
            "%s%s", (is_common_path(path) ? "" : FILESYSTEM_PATH), path);

  return new_path;
}


static void
try_native_bin_cat(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/bin/cat", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_bin_chmod(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/bin/chmod", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_bin_cp(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/bin/cp", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_bin_date(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/bin/date", path))
    return;

  execve(path, argv, envp);
}


static void
try_native_bin_echo(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/bin/echo", path))
    return;

  execve(path, argv, envp);
}


static void
try_native_bin_expr(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/bin/expr", path))
    return;

  execve(path, argv, envp);
}


static void
try_native_bin_hostname(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/bin/hostname", path))
    return;

  execve(path, argv, envp);
}


static void
try_native_bin_ln(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/bin/ln", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_bin_ls(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/bin/ls", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "1adgilnrstACL")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case '1':
        case 'a':
        case 'd':
        case 'g':
        case 'i':
        case 'l':
        case 'n':
        case 'r':
        case 's':
        case 't':
        case 'A':
        case 'C':
        case 'L':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_bin_mkdir(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/bin/mkdir", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_bin_mv(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/bin/mv", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_bin_pwd(const char *path, int argc, char **argv, char * const *envp)
{
  char *p;

  if (strcmp("/bin/pwd", path))
    return;

  if (argc != 1)
    return;

  p = getcwd(NULL, 0);
  if (p == NULL)
    return;

  fprintf(stdout, "%s\n", host_to_target_path(p)); fflush(stdout);

  free(p);

  exit(0);
}


static void
try_native_bin_rm(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/bin/rm", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_bin_rmdir(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/bin/rmdir", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_bin_sh(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/bin/sh", path))
    return;

  /* "/bin/sh -c true" is common enough to special case. */
  if (argc == 3 && !strcmp("-c", argv[1]) && !strcmp("true", argv[2]))
    exit(0);

  if (setenv("EMUL_HW_MACHINE", emul_hw_machine, 1) < 0)
    xfatal("No memory!\n");

  execve(SIMPREFIX "/bin/nbsimsh-arm", argv, envp);
}


static void
try_native_bin_sleep(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/bin/sleep", path))
    return;

  /* Yeah, it is silly ro run this natively, but it is called enough times
   * that it annoying to have it in the logs. */
  execve(path, argv, envp);
}


static void
try_native_usr_bin_ar(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/usr/bin/ar", path))
    return;

  if (argc < 2)
    return;

  for (i = 2; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-ar", argv, envp);
}


static void
try_native_usr_bin_as(const char *path, int argc, char **argv, char * const *envp)
{
  int i = 1;

  if (strcmp("/usr/bin/as", path))
    return;

  while (argv[i] != NULL)
    {
      if (!strcmp(argv[i], "-o") && argv[i + 1] != NULL)
        {
          argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
          i += 2;
        }
      else if (!strncmp(argv[i], "-k", 2))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-m", 2))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "--version", 2))
        {
          i += 1;
        }
      else if (argv[i][0] == '-')
        {
          /* Unknown option. */
          return;
        }
      else
        {
          /* This is a file name. */
          argv[i] = xstrdup(target_to_host_path(argv[i]));
          i++;
        }
    }

  execve(TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-as", argv, envp);
}


static void
try_native_usr_bin_awk(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/awk", path))
    return;

  if (argc == 1 || argc == 2)
    execve(path, argv, envp);

  /* XXX Should check that the program is safe. */
  if (argc == 3 && argv[1][0] != '-' && argv[2][0] != '-')
    {
      argv[2] = xstrdup(target_to_host_path(argv[2]));
      execve(path, argv, envp);
    }

  /* XXX Should check that the program is safe. */
  if (argc == 3 && (!strcmp(argv[1], "--") || !strncmp(argv[1], "-F", 2)))
    execve(path, argv, envp);

  if (argc == 5
      && !strcmp(argv[1], "-f")
      && strstr(argv[2], "mk/bulk/tflat")
      && (!strcmp(argv[3], "-u") || !strcmp(argv[3], "up") ||
          !strcmp(argv[3], "-d") || !strcmp(argv[3], "down"))
      && argv[4][0] != '-')
    {
      argv[2] = xstrdup(target_to_host_path(argv[2]));
      argv[4] = xstrdup(target_to_host_path(argv[4]));

      execve(path, argv, envp);
    }

  if (argc == 5
      && !strcmp(argv[1], "-F")
      && !strcmp(argv[2], "'")
      && !strcmp(argv[3], "/^PACKAGE_VERSION=/ {print $2}"))
    {
      argv[4] = xstrdup(target_to_host_path(argv[4]));

      execve(path, argv, envp);      
    }
}



static void
try_native_usr_bin_basename(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/basename", path))
    return;

  /* No need to transform tha path, since basenames purpose is to remove
   * it anyway... */
  execve(path, argv, envp);
}


static void
try_native_usr_bin_bzcat(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/bzcat", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "s")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 's':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_bzip2(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/bzip2", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "123456789f")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case 'f':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_cc(const char *path, int argc, char **xargv, char * const *envp)
{
  int i;
  int linking = 1;
  int is_shared = 0;

  char **argv;
  int is_cplusplus = 0;
  int is_f77 = 0;

  if (argc == 2 && !strcmp("-v", xargv[1]))
    execve(path, xargv, envp);
  if (argc == 2 && !strcmp("-dumpversion", xargv[1]))
    execve(path, xargv, envp);

  if (!strcmp("/usr/bin/cc", path) ||
      !strcmp("/usr/bin/gcc", path))
    {
      path = TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-gcc";
      /* XXX Need to recognize C++ file prefix. But observe
       * that it does not add -libstdc++! */
    }
  else if (!strcmp("/usr/bin/c++", path))
    {
      path = TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-c++";

      is_cplusplus = 1;
    }
  else if (!strcmp("/usr/bin/g++", path))
    {
      path = TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-g++";

      is_cplusplus = 1;
    }
  else if (!strcmp("/usr/bin/f77", path) ||
           !strcmp("/usr/bin/g77", path))
    {
      path = TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-g77";

      is_f77 = 1;
    }
  else
    return;

  /* cpp seems to be broken.  It need to treat the C++ system files somewhat
   * differently from C files, but there does not seem to be a way to
   * tell it that our -isystem is for C++, so some packages breaks
   * because the include files are treated as C.  grr.
   *
   * Run the pre-processor emulated for C++.  It is rather light-weight, and
   * we will catch the actual compiler (cc1plus) later anyway. */
  if (is_cplusplus)
    return;

  // XXX the array size is never checked below.
  argv = xmalloc((ARGMAX + 1) * sizeof(char*));

  for (i = 0; xargv[i] != NULL; i++)
    argv[i] = xargv[i];
  argv[i] = NULL;

  i = 1;
  while (argv[i] != NULL && i < ARGMAX)
    {
      if (!strcmp(argv[i], "-ansi"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-c"))
        {
          i++;
          linking = 0;
        }
      else if (!strcmp(argv[i], "-dM"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-f") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-f", 2))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-g", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-l") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-l", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-malign-double"))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-march=", 7))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-mcpu=", 6))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-mtune=", 7))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-o") && argv[i + 1] != NULL)
        {
          argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
          i += 2;
        }
      else if (!strncmp(argv[i], "-o", 2))
        {
          char *tmp;
          xasprintf(&tmp, "-o%s", target_to_host_path(argv[i] + 2));
          argv[i++] = tmp;
        }
      else if (!strcmp(argv[i], "-pedantic"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-pedantic-errors"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-p"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-pg"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-pipe"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-shared"))
        {
          is_shared = 1;
          i += 1;
        }
      else if (!strcmp(argv[i], "-static"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-traditional"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-traditional-cpp"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "--version"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "--whole-archive"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-C"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-D") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-D", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-E"))
        {
          i++;
          linking = 0;
        }
      else if (!strcmp(argv[i], "-idirafter") && argv[i + 1] != NULL)
        {
          argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
          i += 2;
        }
      else if (!strcmp(argv[i], "-I"))
        {
          argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
          i += 2;
        }
      else if (!strncmp(argv[i], "-I", 2))
        {
          char *tmp;
          xasprintf(&tmp, "-I%s", target_to_host_path(argv[i] + 2));
          argv[i++] = tmp;
        }
      else if (!strcmp(argv[i], "-L") && argv[i + 1] != NULL)
        {
          argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
          i += 2;
        }
      else if (!strncmp(argv[i], "-L", 2))
        {
          char *tmp;
          xasprintf(&tmp, "-L%s", target_to_host_path(argv[i] + 2));
          argv[i++] = tmp;
        }
      else if (!strncmp(argv[i], "-M", 2))
        {
          i++;
        }
      else if (!strncmp(argv[i], "-O", 2))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-s"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-S"))
        {
          i++;
          linking = 0;
        }
      else if (!strcmp(argv[i], "-U") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-U", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-w"))
        {
          i++;
        }
#if 0
      else if (!strcmp(argv[i], "-W") && argv[i + 1] != NULL)
        {
          i += 2;
        }
#endif
      else if (!strcmp(argv[i], "-Wp,-lang-c89"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-Wl,-E"))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-Wl,-R", 6))
        {
          /* XXX This is only true if the arg is a directory. NOT if it
           * is a file!!! */
          i += 1;
        }
      else if (!strncmp(argv[i], "-Wl,--rpath,", 12))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-Wl,-x"))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-W", 2))
        {
          if (!strncmp(argv[i], "-Wp,-MD,", 8)
              && strncmp(argv[i], "-Wp,-MD,/", 9))
            {
              i += 1;
            }
          else if (!strncmp(argv[i], "-Wl,", 4)
                   || !strncmp(argv[i], "-Wa,", 4)
                   || !strncmp(argv[i], "-Wp,", 4))
            {
              /* Unknown option. */
              return;
            }
          else
            {
              i += 1;
            }
        }
      else if (!strcmp(argv[i], "-x") && argv[i + 1] != NULL)
        {
          if (!strcmp(argv[i + 1], "c"))
            {
              is_cplusplus = 0;
              is_f77 = 0;
            }
          else if (!strcmp(argv[i + 1], "c++"))
            {
              is_cplusplus = 1;
              is_f77 = 0;
            }
          else
            {
              /* XXX Need to add the other valid languages. */
              return;
            }
          i += 2;
        }
      else if (!strncmp(argv[i], "-x", 2))
        {
          if (!strcmp(argv[i], "-xc"))
            {
              is_cplusplus = 0;
              is_f77 = 0;
            }
          else if (!strcmp(argv[i], "-xc++"))
            {
              is_cplusplus = 1;
              is_f77 = 0;
            }
          else
            {
              /* XXX Need to add the other valid languages. */
              return;
            }
          i += 1;
        }
      else if (!strcmp(argv[i], "-"))
        {
          i += 1;
        }
      else if (argv[i][0] == '-')
        {
          /* Unknown option. */
          return;
        }
      else
        {
          /* This is a file name. */
          argv[i] = xstrdup(target_to_host_path(argv[i]));

          /* Look at the file name to make sure the C++ driver is chosen
           * for C++ files, even if the compiler was invoked as gcc. */
          /* XXX This should be done in a way equivalent to what gcc does. */
          if (!is_cplusplus)
            {
              int namelen = strlen(argv[i]);
              is_cplusplus = ((namelen > 4 &&
                               argv[i][namelen - 1] == 'p' &&
                               argv[i][namelen - 2] == 'p' &&
                               argv[i][namelen - 3] == 'c' &&
                               argv[i][namelen - 4] == '.') ||
                              (namelen > 3 &&
                               argv[i][namelen - 1] == 'c' &&
                               argv[i][namelen - 2] == 'c' &&
                               argv[i][namelen - 3] == '.') ||
                              (namelen > 2 &&
                               argv[i][namelen - 1] == 'C' &&
                               argv[i][namelen - 2] == '.'));
            }

          i++;
        }
    }

  if (linking || is_shared)
    {
      int k;
      int j = 1;
      // XXX the array size is never checked below.
      char **tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));

      tmp_argv[0] = argv[0];
      tmp_argv[j++] = "-nostdlib";
      if (is_shared)
        {
          tmp_argv[j++] = "-shared";
          tmp_argv[j++] = FILESYSTEM_PATH "/usr/lib/crtbeginS.o";
        }
      else
        {
          tmp_argv[j++] = "-Wl,-e__start";
          tmp_argv[j++] = FILESYSTEM_PATH "/usr/lib/crt0.o";
          tmp_argv[j++] = FILESYSTEM_PATH "/usr/lib/crtbegin.o";
        }

      for (k = 1; k < i; k++)
        {
          tmp_argv[j++] = argv[k];
          if (!strncmp(argv[k], "-Wl,-R", 6))
            xasprintf(&tmp_argv[j++], "-Wl,-rpath-link,%s",
                      target_to_host_path(&argv[k][6]));
          else if (!strncmp(argv[k], "-Wl,--rpath,", 12))
            xasprintf(&tmp_argv[j++], "-Wl,-rpath-link,%s",
                      target_to_host_path(&argv[k][12]));
              
        }
      argv = tmp_argv;
      i = j;
    }

  argv[i++] = "-nostdinc";
  if (is_cplusplus)
    {
      argv[i++] = "-isystem";
      argv[i++] = FILESYSTEM_PATH "/usr/include/g++";
    }
  argv[i++] = "-isystem";
  argv[i++] = FILESYSTEM_PATH "/usr/include";

  if (linking || is_shared)
    {
      if (is_cplusplus)
        {
          argv[i++] = "-lstdc++";
          argv[i++] = "-lm";
        }
      else if (is_f77)
        {
          argv[i++] = "-lg2c";
          argv[i++] = "-lm";
        }

      if (is_shared)
        {
          argv[i++] = "-lgcc_pic";
          argv[i++] = "-lgcc_pic";
          argv[i++] = FILESYSTEM_PATH "/usr/lib/crtendS.o";
        }
      else
        {
          argv[i++] = "-lgcc"; 
          argv[i++] = "-lc";
          argv[i++] = "-lgcc";
          argv[i++] = FILESYSTEM_PATH "/usr/lib/crtend.o";
        }
      argv[i++] = "-Wl,-rpath-link," FILESYSTEM_PATH "/usr/lib";
      argv[i++] = "-L" FILESYSTEM_PATH "/usr/lib";
    }

  argv[i++] = NULL;

  execve(path, argv, envp);
}


static void
try_native_usr_bin_cmp(const char *path, int argc, char **argv, char * const *envp)
{
  int i = 1;

  if (strcmp("/usr/bin/cmp", path))
    return;

  if (argc == 4 && !strcmp(argv[1], "-s"))
    i++;
  else if (argc != 3)
    return;

  if ((argv[i][0] == '-' && argv[i][1] != '\0') ||
      (argv[i + 1][0] == '-' && argv[i + 1][1] == '\0'))
    return;

  argv[i] = xstrdup(target_to_host_path(argv[i]));
  argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));

  execve(path, argv, envp);
}


static void
try_native_usr_bin_col(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/col", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "bfl:px")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'b':
        case 'f':
        case 'p':
        case 'x':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 'l':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  /* col(1) does not have any plain arguments. */
  tmp_argv[i++] = NULL;
  if (argv[my_optind] != NULL)
    {
      goto out;
    }

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_comm(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/comm", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "123f")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case '1':
        case '2':
        case '3':
        case 'f':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_compress(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/compress", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "b:cfv")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'c':
        case 'f':
        case 'v':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 'b':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_cpp(const char *path, int argc, char **xargv, char * const *envp)
{
  int i;

  char **argv;
  int is_cplusplus = 0;  // XXX

  if (strcmp("/usr/bin/cpp", path))
    return;

  // XXX the array size is never checked below.
  argv = xmalloc((ARGMAX + 1) * sizeof(char*));

  for (i = 0; xargv[i] != NULL; i++)
    argv[i] = xargv[i];
  argv[i] = NULL;

  i = 1;
  while (argv[i] != NULL && i < ARGMAX)
    {
      if (!strcmp(argv[i], "-traditional"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-C"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-D") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-D", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-I"))
        {
          argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
          i += 2;
        }
      else if (!strncmp(argv[i], "-I", 2))
        {
          char *tmp;
          xasprintf(&tmp, "-I%s", target_to_host_path(argv[i] + 2));
          argv[i++] = tmp;
        }
      else if (!strcmp(argv[i], "-M"))
        {
          i++;
        }
      else if (!strncmp(argv[i], "-P", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-undef"))
        {
          i++;
        }
      else if (!strcmp(argv[i], "-U") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-U", 2))
        {
          i += 1;
        }
      else if (argv[i][0] == '-')
        {
          /* Unknown option. */
          return;
        }
      else
        {
          /* This is a file name. */
          argv[i] = xstrdup(target_to_host_path(argv[i]));

          i++;
        }
    }

  argv[i++] = "-nostdinc";
  if (is_cplusplus)
    {
      argv[i++] = "-isystem";
      argv[i++] = FILESYSTEM_PATH "/usr/include/g++";
    }
  argv[i++] = "-isystem";
  argv[i++] = FILESYSTEM_PATH "/usr/include";

  argv[i++] = NULL;

  execve(TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-cpp", argv, envp);
}

static void
try_native_usr_bin_cut(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/cut", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "b:c:d:f:ns")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'n':
        case 's':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 'b':
        case 'c':
        case 'd':
        case 'f':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_dc(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/dc", path))
    return;

  if (argc == 1)
    execve(path, argv, envp);
}


static void
try_native_usr_bin_diff(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/diff", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "vC:")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'v':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 'C':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_env(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  char **tmp_envp;
  char *v;

  if (strcmp("/usr/bin/env", path))
    return;

  tmp_envp = xmalloc(sizeof(char*) * ENVMAX);

  i = 1;
  if (argv[i] != NULL && !strcmp(argv[i], "-i"))
    {
      i++;
      tmp_envp[0] = NULL;
    }
  else
    {
      int j;

      for (j = 0; envp[j] != NULL; j++)
        tmp_envp[j] = envp[j];
      tmp_envp[j] = NULL;
    }


  for (; i < argc && (v = strchr(argv[i], '=')); i++)
    {
      char **tmp;

      for (tmp = tmp_envp; *tmp != NULL; tmp++)
        if (!strncmp(*tmp, argv[i], v - argv[i] + 1))
          break;

      *tmp = argv[i];
    }

  if (argv[i] == NULL)
    {
      while (*tmp_envp != 0)
        printf("%s\n", *tmp_envp++);

      exit(0);
    }

  /* XXX This should emulate execvp() instead... */
  run_execve(argv[i], &argv[i], tmp_envp);

  /* Errno is set to something if we reach this point, which will make
   * the emulator think that the execution of env failed (but the error
   * is probebly due to the difference between execve() and execvp().
   *
   * Clear errno to let the command be run emulated. */
  errno = 0;
}


static void
try_native_usr_bin_eqn(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/usr/bin/eqn", path))
    return;

  i = 1;
  while (argv[i] != NULL)
    {
      if (!strcmp(argv[i], "-Tascii"))
        {
          i += 1;
        }
      else
        {
          /* Unknown option. */
          return;
        }
    }

  execve(path, argv, envp);
}


static void
try_native_usr_bin_false(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/false", path))
    return;

  exit(1);
}


static void
try_native_usr_bin_file(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/file", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "bnfL")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'b':
        case 'n':
        case 'L':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 'f':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_find(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/find", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "df:sxX")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
#if 0 /* XXX: Until tested. */
        case 'd':
        case 's':
        case 'x':
        case 'X':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;
#endif
        case 'f':
          xasprintf(&tmp_argv[i++],
                    "-%c%s", ch, target_to_host_path(my_optarg));
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    {
      if (!strcmp(argv[my_optind], "-newer")
          && argv[my_optind + 1] != NULL)
        {
          tmp_argv[i++] = xstrdup(argv[my_optind++]);
          tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
        }
      else if (!strcmp(argv[my_optind], "-type")
               && argv[my_optind + 1] != NULL)
        {
          tmp_argv[i++] = xstrdup(argv[my_optind++]);
          tmp_argv[i++] = xstrdup(argv[my_optind++]);
        }
      else if (!strcmp(argv[my_optind], "-print"))
        {
          tmp_argv[i++] = xstrdup(argv[my_optind++]);
        }
      else if (argv[my_optind][0] == '-')
        {
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
      else
        {
          if (!is_common_path(argv[my_optind]))
            {
              /* We cannot allow file names that need transformation,
               * since the output will then also be transformed,
               * which may cause problmems... */
              tmp_argv[i++] = NULL;
              goto out;
            }

          tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
        }
    }

  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_fold(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/fold", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "bsw:")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'b':
        case 's':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 'w':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_ftp(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/ftp", path))
    return;

  if (argc == 2)
    execve(path, argv, envp);
}


static void
try_native_usr_bin_dirname(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/usr/bin/dirname", path))
    return;

  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_usr_bin_grep(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;
  int got_pattern = 0;

  if (strcmp("/usr/bin/grep", path)
      && strcmp("/usr/bin/egrep", path)
      && strcmp("/usr/bin/fgrep", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "ce:fhilnqrsvwxyA:EF")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'c':
        case 'h':
        case 'i':
        case 'l':
        case 'n':
        case 'q':
        case 'r':
        case 's':
        case 'v':
        case 'w':
        case 'x':
        case 'y':
        case 'E':
        case 'F':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 'e':
          got_pattern = 1;
          /* FALLTHROUGH */
        case 'A':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        case 'f':
          got_pattern = 1;
          /* XXX should not be dup? */
          xasprintf(&tmp_argv[i++],
                    "-%c%s", ch, xstrdup(target_to_host_path(my_optarg)));
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  if (argv[my_optind] != NULL && !got_pattern && i < ARGMAX)
    tmp_argv[i++] = xstrdup(argv[my_optind++]);

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_groff(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/groff", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "m:pstRT:")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'p':
        case 's':
        case 't':
        case 'R':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 'm':
        case 'T':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_grotty(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/grotty", path))
    return;

  if (argc != 1)
    return;

  execve(path, argv, envp);
}


static void
try_native_usr_bin_head(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;
  int j;

  if (strcmp("/usr/bin/head", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  /* Take care of historical args. */
  for (j = 1; j < argc && argv[j][0] == '-' && argv[j][1] != '-'; j++)
    if (isdigit((unsigned char)argv[j][1]))
      {
        char *tmp;

        xasprintf(&tmp, "-n%s", &argv[j][1]);
        argv[j] = tmp;
      }

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "n:")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'n':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_id(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/id", path))
    return;

  if (argc == 2 && !strcmp(argv[1], "-u"))
    execve(path, argv, envp);
  if (argc == 3 && !strcmp(argv[1], "-u") && !strcmp(argv[2], "root"))
    execve(path, argv, envp);
}


static void
try_native_usr_bin_install(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;
  int stripping = 0;

  if (strcmp("/usr/bin/install", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "cdg:l:m:o:prs")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'c':
        case 'd':
        case 'p':
        case 'r':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 's':
          xasprintf(&tmp_argv[i++], "-%c", ch);

          stripping = 1;
          break;

        case 'g':
          /* We must check that the group name does exist on the host
           * before running this native (to prevent problems when
           * a new group has been added in the target). */
          if (strcmp("wheel", my_optarg)
              && strcmp("bin", my_optarg)
              && strcmp("games", my_optarg)
              && strcmp("daemon", my_optarg))
            {
              tmp_argv[i++] = NULL;
              goto out;
            }

        case 'l':
        case 'm':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        case 'o':
          /* We must check that the owner name does exist on the host
           * before running this native (to prevent problems when
           * a new user has been added in the target). */
          if (strcmp("root", my_optarg)
              && strcmp("bin", my_optarg)
              && strcmp("games", my_optarg)
              && strcmp("daemon", my_optarg))
            {
              tmp_argv[i++] = NULL;
              goto out;
            }

          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i >= ARGMAX)
    goto out;

  if (stripping)
    {
      /* We need to set the STRIP env before calling "install -s"
       * natively, but we should not set it in case it is already
       * set to something else... */
      int count;
      char **tmp_envp;

      for (count = 0; envp[count] != NULL; count++)
        if (!strncmp("STRIP=", envp[count], 6))
          goto out;

      tmp_envp = xmalloc((count + 2) * sizeof(char*));

      for (i = 0; i < count; i++)
        tmp_envp[i] = envp[i];
      xasprintf(&tmp_envp[i++],
            "STRIP=" TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-strip");
      tmp_envp[i] = NULL;

      envp = tmp_envp;
    }

  execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_ld(const char *path, int argc, char **xargv, char * const *envp)
{
  int i;
  char **argv;

  if (strcmp("/usr/bin/ld", path))
    return;

  // XXX the array size is never checked below.
  argv = xmalloc((ARGMAX + 1) * sizeof(char*));

  for (i = 0; xargv[i] != NULL; i++)
    argv[i] = xargv[i];
  argv[i] = NULL;

  i = 1;
  while (argv[i] != NULL)
    {
      if (!strcmp(argv[i], "-dynamic-linker") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strcmp(argv[i], "-e") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strcmp(argv[i], "-E"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-export-dynamic"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-d")
               || !strcmp(argv[i], "-dc")
               || !strcmp(argv[i], "-dp"))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-l", 2) && strlen(argv[i]) > 2)
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-L") && argv[i + 1] != NULL)
        {
          argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
          i += 2;
        }
      else if (!strncmp(argv[i], "-L", 2))
        {
          char *tmp;
          xasprintf(&tmp, "-L%s", target_to_host_path(argv[i] + 2));
          argv[i++] = tmp;
        }
      else if (!strcmp(argv[i], "-r"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-o") && argv[i + 1] != NULL)
        {
          argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
          i += 2;
        }
      else if (!strcmp(argv[i], "-R") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-R", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-shared"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-v"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-x"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-X"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "--help"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "--version"))
        {
          i += 1;
        }
      else if (argv[i][0] == '-')
        {
          /* Unknown option. */
          return;
        }
      else
        {
          /* This is a file. */
          argv[i] = xstrdup(target_to_host_path(argv[i]));
          i += 1;
        }
    }

  argv[i++] = "-L";
  argv[i++] = FILESYSTEM_PATH "/usr/lib";
  argv[i++] = "-rpath-link";
  argv[i++] = FILESYSTEM_PATH "/usr/lib";
  argv[i++] = NULL;

  execve(TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-ld", argv, envp);
}


static void
try_native_usr_bin_make(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/make", path))
    return;

  if (setenv("EMUL_HW_MACHINE", emul_hw_machine, 1) < 0)
    xfatal("No memory!\n");

  execve(SIMPREFIX "/bin/nbsimmake-arm", argv, envp);
}


static void
try_native_usr_bin_makeinfo(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/makeinfo", path))
    return;

  if (argc == 2 && !strcmp("--version", argv[1]))
    execve(path, argv, envp);
}


static void
try_native_usr_bin_gunzip(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/gunzip", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "cfd")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'c':
        case 'f':
        case 'd':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_gzcat(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/gzcat", path))
    return;

  if (argc == 1)
    execve(path, argv, envp);

  if (argc == 2 && argv[1][0] != '-')
    {
      argv[1] = xstrdup(target_to_host_path(argv[1]));
      execve(path, argv, envp);
    }
}


static void
try_native_usr_bin_gzip(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/usr/bin/gzip", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_usr_bin_mktemp(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/mktemp", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "dq")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'd':
        case 'q':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    if (is_common_path(argv[my_optind]))
      tmp_argv[i++] = xstrdup(argv[my_optind++]);
    else
      {
        /* XXX We need to transform the output after this call.
         * Just run it emulated for now. */
        tmp_argv[i++] = NULL;
        goto out;
      }
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_nm(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/nm", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "gBV")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'g':
        case 'B':
        case 'V':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-nm",
           tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_nroff(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/nroff", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "him:n:o:r:T:")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'h':
        case 'i':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 'm':
        case 'n':
        case 'o':
        case 'r':
        case 'T':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_od(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/od", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "acoxB")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'a':
        case 'c':
        case 'o':
        case 'x':
        case 'B':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_patch(const char *path, int argc, char **argv, char * const *envp)
{
  int i = 1;

  if (strcmp("/usr/bin/patch", path))
    return;

  while (argv[i] != NULL)
    {
      if (!strcmp(argv[i], "-b") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-b", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "--batch"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-d") && argv[i + 1] != NULL)
        {
          argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
          i += 2;
        }
      else if (!strncmp(argv[i], "-d", 2))
        {
          char *tmp;
          xasprintf(&tmp, "-d%s", target_to_host_path(argv[i] + 2));
          argv[i++] = tmp;
        }
      else if (!strcmp(argv[i], "--forward"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-p") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-p", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "--quiet"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-v"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-E"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-F") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-F", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-V") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-V", 2))
        {
          i += 1;
        }
      else
        {
          /* Unknown option. */
          return;
        }
    }

  execve(path, argv, envp);
}


static void
try_native_usr_bin_ranlib(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/ranlib", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "vV")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'v':
        case 'V':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-ranlib",
           tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_sed(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/usr/bin/sed", path))
    return;

  if (argc < 2)
    return;


  if (argv[1][0] != '-')
    {
      /* sed command [file ...] */
      i = 2;
    }
  else
    {
      int got_command = 0;

      /* "sed [-n] [-e command] [-f command_file] [file ...]" */
      i = 1;

      while (argv[i] != NULL)
        {
          if (!strcmp(argv[i], "-e") && argv[i + 1] != NULL)
            {
              i += 2;
              got_command = 1;
            }
          else if (!strcmp(argv[i], "-ne") && argv[i + 1] != NULL) /* XXX */
            {
              i += 2;
              got_command = 1;
            }
          else if (!strcmp(argv[i], "-f") && argv[i + 1] != NULL)
            {
              argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
              i += 2;
              got_command = 1;
            }
          else if (!strcmp(argv[i], "-n"))
            i += 1;
          else if (argv[i][0] != '-')
            {
              /* The rest of the arguments are files. */
              break;
            }
          else
            {
              /* We don't handle these arguments. */
              return;
            }
        }

      if (!got_command && argv[i] != NULL)
        i++;
    }

  while (argv[i] != NULL)
    {
      argv[i] = xstrdup(target_to_host_path(argv[i]));
      i++;
    }

  execve(path, argv, envp);
}


static void
try_native_usr_bin_sort(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/sort", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "bcdfik:mHno:rR:sSt:uT:")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'b':
        case 'c':
        case 'd':
        case 'f':
        case 'i':
        case 'm':
        case 'H':
        case 'n':
        case 'r':
        case 's':
        case 'S':
        case 'u':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        case 'k':
        case 'R':
        case 't':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        case 'o':
        case 'T':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, target_to_host_path(my_optarg));
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_strip(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/strip", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "sV")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 's':
        case 'V':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(TOOL_PATH "/bin/" EMUL_HW_MACHINE_ARCH "--netbsdelf-strip",
           tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_su(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/su", path))
    return;

  if (argc == 5
      && !strcmp(argv[1], "-")
      && !strcmp(argv[2], "root")
      && !strcmp(argv[3], "-c"))
    {
      /* pkgsrc calls su to run the command 
       *    cd path ; env nbsimmake real-su-install
       * which is safe to run natively (since it jups back into the emulator
       * when nbsimmake is called). */
      char *s;

      s = strchr(argv[4], ';');

      if (!strncmp(argv[4], "cd ", 3)
          && s != NULL
          && !strncmp(s, "; /usr/bin/env ", 15)
          && strchr(s + 1, ';') == NULL) /* XXX Need a real parser. */
        {
          char* p1 = xstrdup(argv[4]);
          char* p2 = strstr(p1, " /usr/bin/make ");
          if (p2 == NULL)
            return;
          if (strstr(p2 + 1, " /usr/bin/make ") != NULL)
            return;

          *p2 = 0;
          p2 += 15;

          xasprintf(&argv[4],
                    "%s EMUL_HW_MACHINE=%s " SIMPREFIX "/bin/nbsimmake-arm %s",
                    p1, emul_hw_machine, p2);

          execve(path, argv, envp);
        }
    }
}


static void
try_native_usr_bin_tail(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;
  int j;

  if (strcmp("/usr/bin/tail", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  /* Take care of historical args. */
  for (j = 1; j < argc && argv[j][0] == '-' && argv[j][1] != '-'; j++)
    if (isdigit((unsigned char)argv[j][1]))
      {
        char *tmp;

        xasprintf(&tmp, "-n%s", &argv[j][1]);
        argv[j] = tmp;
      }

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "n:")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'n':
          xasprintf(&tmp_argv[i++], "-%c%s", ch, my_optarg);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_tbl(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/tbl", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "Cv")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'C':
        case 'v':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_tee(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;

  if (strcmp("/usr/bin/tee", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "ai")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'a':
        case 'i':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_touch(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/usr/bin/touch", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_usr_bin_tr(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/tr", path))
    return;

  execve(path, argv, envp);
}


static void
try_native_usr_bin_true(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/true", path))
    return;

  exit(0);
}


static void
try_native_usr_bin_tsort(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/usr/bin/tsort", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_usr_bin_uname(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/uname", path))
    return;

  if (argc == 2 && !strcmp("-p", argv[1]))
    {
      fprintf(stdout, "%s\n", EMUL_HW_MACHINE_ARCH); fflush(stdout);
      exit(0);
    }
  if (argc == 2 && !strcmp("-m", argv[1]))
    {
      fprintf(stdout, "%s\n", emul_hw_machine); fflush(stdout);
      exit(0);
    }
  if (argc == 2 && !strcmp("-r", argv[1]))
    {
      fprintf(stdout, "%s\n", EMUL_KERN_OSRELEASE); fflush(stdout);
      exit(0);
    }
  if (argc == 1 ||
      (argc == 2 && !strcmp("-s", argv[1])))
    {
      fprintf(stdout, "NetBSD\n"); fflush(stdout);
      exit(0);
    }

  if (argc == 2 && !strcmp("-n", argv[1]))
    execve(path, argv, envp);
}


static void
try_native_usr_bin_uniq(const char *path, int argc, char **argv, char * const *envp)
{
  char** tmp_argv;
  int ch;
  int i = 0;
  int j;

  if (strcmp("/usr/bin/uniq", path))
    return;

  tmp_argv = xmalloc((ARGMAX + 1) * sizeof(char*));
  tmp_argv[i++] = xstrdup(argv[0]);

  /* Take care of historical args. */
  for (j = 1; j < argc && argv[j][0] == '-' && argv[j][1] != '-'; j++)
    if (isdigit((unsigned char)argv[j][1]))
      {
        char *tmp;

        xasprintf(&tmp, "-f%s", &argv[j][1]);
        argv[j] = tmp;
      }

  my_optreset = my_optind = 1;
  while ((ch = my_getopt(argc, argv, "cdu")) != -1
         && i < ARGMAX)
    {
      switch (ch)
        {
        case 'c':
        case 'd':
        case 'u':
          xasprintf(&tmp_argv[i++], "-%c", ch);
          break;

        default:
          /* Unknown option. */
          tmp_argv[i++] = NULL;
          goto out;
        }
    }

  while (argv[my_optind] != NULL && i < ARGMAX)
    tmp_argv[i++] = xstrdup(target_to_host_path(argv[my_optind++]));
  tmp_argv[i++] = NULL;

  if (i < ARGMAX)
    execve(path, tmp_argv, envp);

 out:
  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
}


static void
try_native_usr_bin_wc(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp("/usr/bin/wc", path))
    return;

  /* It is OK to transform all arguments, since only arguments
   * starting with '/' will be modified. */
  for (i = 1; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static void
try_native_usr_bin_whoami(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/whoami", path))
    return;

  execve(path, argv, envp);
}


static void
try_native_usr_bin_yes(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/bin/yes", path))
    return;

  execve(path, argv, envp);
}


static void
try_native_sbin_sysctl(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/sbin/sysctl", path))
    return;

  if (argc == 3 &&
      !strcmp(argv[1], "-n") && !strcmp(argv[2], "hw.machine_arch"))
    {
      fprintf(stdout, "%s\n", EMUL_HW_MACHINE_ARCH);
      exit(0);
    }
  if (argc == 3 &&
      !strcmp(argv[1], "-n") && !strcmp(argv[2], "hw.model"))
    {
      fprintf(stdout, "%s\n", EMUL_HW_MODEL);
      exit(0);
    }

  /* XXX Need to list "safe" sysctls. */
  if (argc == 3 &&
      !strcmp(argv[1], "-n") && !strcmp(argv[2], "kern.argmax"))
    {
      execve(path, argv, envp);
    }
}


static void
try_native_usr_sbin_pkg_admin(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/sbin/pkg_admin", path))
    return;

  /* pmatch is just a glorified regexp, so we can do it natively. */
  if (argc == 4
      && !strcmp("pmatch", argv[1])
      && argv[2][0] != '-'
      && argv[3][0] != '-')
    execve(path, argv, envp);
  if (argc == 6
      && !strcmp("-K", argv[1])
      && !strcmp("pmatch", argv[3])
      && argv[4][0] != '-'
      && argv[5][0] != '-')
    {
      argv[2] = xstrdup(target_to_host_path(argv[2]));
      execve(path, argv, envp);
    }
}


static void
try_native_usr_sbin_pkg_info(const char *path, int argc, char **argv, char * const *envp)
{
  if (strcmp("/usr/sbin/pkg_info", path))
    return;

  if (argc == 2 && !strcmp("-V", argv[1]))
    {
      execve(path, argv, envp);
    }

  if (argc == 3
      && (!strcmp("-e", argv[1]) || !strcmp("-qe", argv[1]))
      && argv[2][0] != '-')
    {
      int count, i, found = 0;
      char **tmp_envp;

      for (count = 0; environ[count] != NULL; count++)
        ;

      tmp_envp = xmalloc((count + 1) * sizeof(char*));

      for (i = 0; i < count; i++)
        {
          if (!strncmp("PKG_DBDIR=", envp[i], 10))
            {
              xasprintf(&tmp_envp[i], "PKG_DBDIR=%s",
                        target_to_host_path(envp[i] + 10));
              found = 1;
            }
          else
            tmp_envp[i] = xstrdup(envp[i]);
        }
      tmp_envp[i] = NULL;

      if (found)
        execve(path, argv, tmp_envp);

      for (i = 0; i < count; i++)
        free(tmp_envp[i]);
      free(tmp_envp);
    }

  if (argc == 5
      && !strcmp("-K", argv[1])
      && (!strcmp("-e", argv[3]) || !strcmp("-qe", argv[3]))
      && argv[4][0] != '-')
    {
      argv[2] = xstrdup(target_to_host_path(argv[2]));

      execve(path, argv, envp);
    }
}


static void
try_native_usr_libexec_cc1(const char *path, int argc, char **argv, char * const *envp)
{
  int i = 1;

  if (!strcmp("/usr/libexec/cc1", path))
    path =
      TOOL_PATH "/lib/gcc-lib/" EMUL_HW_MACHINE_ARCH "--netbsdelf/2.95.3/cc1";
  else   if (!strcmp("/usr/libexec/cc1plus", path))
    path =
      TOOL_PATH "/lib/gcc-lib/" EMUL_HW_MACHINE_ARCH \
      "--netbsdelf/2.95.3/cc1plus";
  else
    return;

  while (argv[i] != NULL)
    {
      if (!strcmp(argv[i], "-ansi"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-dumpbase") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strcmp(argv[i], "-f") && argv[i + 1] != NULL)
        {
          i += 2;
        }
      else if (!strncmp(argv[i], "-f", 2))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-g", 2))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-m", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-p") || !strcmp(argv[i], "-pg"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-o") && argv[i + 1] != NULL)
        {
          argv[i + 1] = xstrdup(target_to_host_path(argv[i + 1]));
          i += 2;
        }
      else if (!strncmp(argv[i], "-O", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-pedantic"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-pedantic-errors"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-quiet"))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-version"))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-w", 2))
        {
          i += 1;
        }
      else if (!strncmp(argv[i], "-W", 2))
        {
          i += 1;
        }
      else if (!strcmp(argv[i], "-") && argv[i] == NULL)
        {
          i += 1;
        }
      else if (argv[i][0] == '-')
        {
          /* Unknown option. */
          return;
        }
      else
        {
          /* This is a file name. */
          argv[i] = xstrdup(target_to_host_path(argv[i]));
          i++;
        }
    }

  execve(path, argv, envp);
}


static void
try_usr_pkg_bin_digest(const char *path, int argc, char **argv, char * const *envp)
{
  int i;

  if (strcmp(PKGPREFIX "/bin/digest", path))
    return;

  for (i = 2; i < argc; i++)
    argv[i] = xstrdup(target_to_host_path(argv[i]));

  execve(path, argv, envp);
}


static int
try_native_execve(const char *path, char * const *argv, char * const *envp)
{
  int i;
  char **tmp_argv;
  int argc;

  for (argc = 0; argv[argc] != NULL; argc++)
    ;

  tmp_argv = xmalloc((argc + 1) * sizeof(*tmp_argv));
  for (i = 0; i < argc + 1; i++)
    tmp_argv[i] = argv[i];

  errno = 0;

  if (!strncmp("/bin/", path, 5))
    {
      switch (path[5])
        {
        case 'c':
          try_native_bin_cat(path, argc, tmp_argv, envp);
          try_native_bin_chmod(path, argc, tmp_argv, envp);
          try_native_bin_cp(path, argc, tmp_argv, envp);
          break;
        case 'd':
          try_native_bin_date(path, argc, tmp_argv, envp);
          break;
        case 'e':
          try_native_bin_echo(path, argc, tmp_argv, envp);
          try_native_bin_expr(path, argc, tmp_argv, envp);
          break;
        case 'h':
          try_native_bin_hostname(path, argc, tmp_argv, envp);
          break;
        case 'l':
          try_native_bin_ln(path, argc, tmp_argv, envp);
          try_native_bin_ls(path, argc, tmp_argv, envp);
          break;
        case 'm':
          try_native_bin_mkdir(path, argc, tmp_argv, envp);
          try_native_bin_mv(path, argc, tmp_argv, envp);
          break;
        case 'p':
          try_native_bin_pwd(path, argc, tmp_argv, envp);
          break;
        case 'r':
          try_native_bin_rm(path, argc, tmp_argv, envp);
          try_native_bin_rmdir(path, argc, tmp_argv, envp);
          break;
        case 's':
          try_native_bin_sh(path, argc, tmp_argv, envp);
          try_native_bin_sleep(path, argc, tmp_argv, envp);
          break;
        default:
          break;
        }
    }
  else if (!strncmp("/usr/bin/", path, 9))
    {
      switch (path[9])
        {
        case 'a':
          try_native_usr_bin_ar(path, argc, tmp_argv, envp);
          try_native_usr_bin_as(path, argc, tmp_argv, envp);
          try_native_usr_bin_awk(path, argc, tmp_argv, envp);
          break;
        case 'b':
          try_native_usr_bin_basename(path, argc, tmp_argv, envp);
          try_native_usr_bin_bzcat(path, argc, tmp_argv, envp);
          try_native_usr_bin_bzip2(path, argc, tmp_argv, envp);
          break;
        case 'c':
          try_native_usr_bin_cc(path, argc, tmp_argv, envp); /* cc, c++ */
          try_native_usr_bin_cmp(path, argc, tmp_argv, envp);
          try_native_usr_bin_col(path, argc, tmp_argv, envp);
          try_native_usr_bin_comm(path, argc, tmp_argv, envp);
          try_native_usr_bin_compress(path, argc, tmp_argv, envp);
          try_native_usr_bin_cpp(path, argc, tmp_argv, envp);
          try_native_usr_bin_cut(path, argc, tmp_argv, envp);
          break;
        case 'd':
          try_native_usr_bin_dc(path, argc, tmp_argv, envp);
          try_native_usr_bin_diff(path, argc, tmp_argv, envp);
          try_native_usr_bin_dirname(path, argc, tmp_argv, envp);
          break;
        case 'e':
          try_native_usr_bin_grep(path, argc, tmp_argv, envp); /* egrep */
          try_native_usr_bin_env(path, argc, tmp_argv, envp);
          try_native_usr_bin_eqn(path, argc, tmp_argv, envp);
          break;
        case 'f':
          try_native_usr_bin_cc(path, argc, tmp_argv, envp); /* f77 */
          try_native_usr_bin_grep(path, argc, tmp_argv, envp); /* fgrep */
          try_native_usr_bin_false(path, argc, tmp_argv, envp);
          try_native_usr_bin_file(path, argc, tmp_argv, envp);
          try_native_usr_bin_find(path, argc, tmp_argv, envp);
          try_native_usr_bin_fold(path, argc, tmp_argv, envp);
          try_native_usr_bin_ftp(path, argc, tmp_argv, envp);
          break;
        case 'g':
          try_native_usr_bin_cc(path, argc, tmp_argv, envp); /* gcc, g++,g77 */
          try_native_usr_bin_grep(path, argc, tmp_argv, envp);
          try_native_usr_bin_groff(path, argc, tmp_argv, envp);
          try_native_usr_bin_grotty(path, argc, tmp_argv, envp);
          try_native_usr_bin_gunzip(path, argc, tmp_argv, envp);
          try_native_usr_bin_gzcat(path, argc, tmp_argv, envp);
          try_native_usr_bin_gzip(path, argc, tmp_argv, envp);
          break;
        case 'h':
          try_native_usr_bin_head(path, argc, tmp_argv, envp);
          break;
        case 'i':
          try_native_usr_bin_id(path, argc, tmp_argv, envp);
          try_native_usr_bin_install(path, argc, tmp_argv, envp);
          break;
        case 'l':
          try_native_usr_bin_ld(path, argc, tmp_argv, envp);
          break;
        case 'm':
          try_native_usr_bin_make(path, argc, tmp_argv, envp);
          try_native_usr_bin_makeinfo(path, argc, tmp_argv, envp);
          try_native_usr_bin_mktemp(path, argc, tmp_argv, envp);
          break;
        case 'n':
          try_native_usr_bin_nm(path, argc, tmp_argv, envp);
          try_native_usr_bin_nroff(path, argc, tmp_argv, envp);
          break;
        case 'o':
          try_native_usr_bin_od(path, argc, tmp_argv, envp);
          break;
        case 'p':
          try_native_usr_bin_patch(path, argc, tmp_argv, envp);
          break;
        case 'r':
          try_native_usr_bin_ranlib(path, argc, tmp_argv, envp);
          break;
        case 's':
          try_native_usr_bin_sed(path, argc, tmp_argv, envp);
          try_native_usr_bin_sort(path, argc, tmp_argv, envp);
          try_native_usr_bin_strip(path, argc, tmp_argv, envp);
          try_native_usr_bin_su(path, argc, tmp_argv, envp);
          break;
        case 't':
          try_native_usr_bin_tail(path, argc, tmp_argv, envp);
          try_native_usr_bin_tbl(path, argc, tmp_argv, envp);
          try_native_usr_bin_tee(path, argc, tmp_argv, envp);
          try_native_usr_bin_touch(path, argc, tmp_argv, envp);
          try_native_usr_bin_tr(path, argc, tmp_argv, envp);
          try_native_usr_bin_true(path, argc, tmp_argv, envp);
          try_native_usr_bin_tsort(path, argc, tmp_argv, envp);
          break;
        case 'u':
          try_native_usr_bin_uname(path, argc, tmp_argv, envp);
          try_native_usr_bin_uniq(path, argc, tmp_argv, envp);
          break;
        case 'w':
          try_native_usr_bin_wc(path, argc, tmp_argv, envp);
          try_native_usr_bin_whoami(path, argc, tmp_argv, envp);
          break;
        case 'y':
          try_native_usr_bin_yes(path, argc, tmp_argv, envp);
          break;
        default:
          break;
        }
    }
  else if (!strncmp("/sbin/", path, 6))
    {
      switch (path[6])
        {
        case 's':
          try_native_sbin_sysctl(path, argc, tmp_argv, envp);
          break;
        default:
          break;
        }
    }
  else
    {
      try_native_usr_libexec_cc1(path, argc, tmp_argv, envp); /* cc1plus */

      try_native_usr_sbin_pkg_admin(path, argc, tmp_argv, envp);
      try_native_usr_sbin_pkg_info(path, argc, tmp_argv, envp);

      try_usr_pkg_bin_digest(path, argc, tmp_argv, envp);
    }

  free(tmp_argv);

  /* If errno indicates an error here, then it mean that we have tried
   * to run a NetBSD base command, and failed (for example due to a
   * too long parameter string -- this happens when configure tries to
   * determine the maximum parameter length).  There is no meaning in
   * trying the same stuff in the emulator... */
  if (errno != 0)
    return -1;

  return 0;
}


int
run_execve(const char *path, char * const *argv, char * const *envp)
{
  int i, status, linklen;
  char **tmp_argv;
  int argc;

  char filebuf[1024]; /* XXX */
  int fd;
  ssize_t len;

  for (argc = 0; argv[argc] != NULL; argc++)
    ;

#if 1
  {
    FILE* f;

    f = fopen("/usr/local/tmp/oexec", "a+");
    if (f != NULL)
      {
        time_t t = time(NULL);
        fprintf(f, "%s:%d:%.24s ", path, argc, ctime(&t));
        for (i = 0; argv[i] != NULL; i++)
          fprintf(f, " %s ", argv[i]);
        fprintf(f, "\n");
        fclose(f);
      }
  }
#endif

  /* This does not return if the program is run natively. */
  status = try_native_execve(path, argv, envp);
  if (status == -1)
    return -1;

  /* Is the path a link to something we can run? */
  linklen = readlink(target_to_host_path(path), filebuf, sizeof(filebuf) - 1);
  if (linklen > 0)
    {
      filebuf[linklen] = '\0';

      /* This does not return if the program is run natively. */
      status = try_native_execve(host_to_target_path(filebuf), argv, envp);
      if (status == -1)
        return -1;
    }

  fd = open(target_to_host_path(path), O_RDONLY, 0);
  if (fd == -1)
    {
      /* XXX need to update errno? */
      return -1;
    }

  len = read(fd, filebuf, sizeof(filebuf));
  close(fd);

  if (len < 2)
    {
      errno = EFAULT;
      return -1;
    }

  i = 0;

  tmp_argv = xmalloc((argc + 10) * sizeof(char*));

  /* Arguments to start the emulator. */
  tmp_argv[i++] = xstrdup("run");
  tmp_argv[i++] = xstrdup("-k");
  tmp_argv[i++] = xstrdup("--");

  if (filebuf[0] == '#' && filebuf[1] == '!')
    {
      int j;
      unsigned char tmp_buf[1024];  // XXX
      unsigned char* tmp_ptr;
      unsigned char* ptr = &filebuf[2];
      while (ptr != '\0' && isspace(*ptr))
        ptr++;
      if (*ptr == '\0')
        {
          errno = ENOENT;
          return -1;
        }

      tmp_ptr = ptr;
      while (*ptr != '\0' && !isspace(*ptr))
        ptr++;
      memcpy(tmp_buf, tmp_ptr, ptr-tmp_ptr);
      tmp_buf[ptr-tmp_ptr] = '\0';

      /* The actual command to run (as seen as the emulator and the
       * emulated progran respectively. */
      tmp_argv[i++] = xstrdup(target_to_host_path(tmp_buf));
      tmp_argv[i++] = xstrdup(tmp_buf);

      /* The command may have one additional argument (rest of line is
       * interpreted as one argument, even if it contains whitespaces). */
      filebuf[sizeof(filebuf) - 1] = '\n';
      while (*ptr != '\n' && isspace(*ptr))
        ptr++;
      if (*ptr != '\n')
        {
          tmp_ptr = ptr;

          while (*ptr != '\n')
            ptr++;
          *ptr = '\0';

          tmp_argv[i++] = xstrdup(tmp_ptr);
        }

      /* The arguments the emulated progam will see. */
      tmp_argv[i++] = xstrdup(path);
      for (j = 1; argv[j] != NULL; j++)
        tmp_argv[i++] = xstrdup(argv[j]);
      tmp_argv[i++] = NULL;

      /* Maybe this new command can be run natively? */
      (void)try_native_execve(tmp_argv[4], tmp_argv + 4, envp);
    }
  else if (filebuf[1] == 'E' && filebuf[2] == 'L' && filebuf[3] == 'F')
    {
      int j;
      tmp_argv[i++] = xstrdup(target_to_host_path(path));

      /* The arguments the emulated progam will see. */
      for (j = 0; argv[j] != NULL; j++)
        tmp_argv[i++] = xstrdup(argv[j]);
      tmp_argv[i++] = NULL;
    }
  else
    {
      errno = ENOEXEC;
      return -1;
    }

#if 1
  {
    FILE* f;

    f = fopen("/usr/local/tmp/exec", "a+");
    if (f != NULL)
      {
        time_t t = time(NULL);
        fprintf(f, "%s:%d:%.24s ", path, argc, ctime(&t));
        for (i = 0; argv[i] != NULL; i++)
          fprintf(f, " %s ", argv[i]);
        fprintf(f, "\n");
        fclose(f);
      }
  }
#endif

  status = execve(SIMPREFIX "/bin/netbsdsim-" EMUL_HW_MACHINE_ARCH,
                  tmp_argv, envp);

  return status;
}


void
netbsd_try_native(const char *path, char * const *argv)
{
  try_native_execve(path, argv, environ);
}
